#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <xenctrl.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <stdarg.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "evtchnd.h"
#include "shared.h"

/* ------------------------------------------------------------------ */

static int libxc_trace = 0;
static int libxc_fixme = 1;

static evtchn_port_t xsd_port;
static int           xsd_fd;
static void          *xsd_map;

/* ------------------------------------------------------------------ */

#include "libxc-evtchn.c"

/* ------------------------------------------------------------------ */

int xc_interface_open(void)
{
    int rc = 42;

    if (libxc_trace)
	fprintf(stderr, "libxc: %s -> handle %d\n", __FUNCTION__, rc);
    return rc;
}

int xc_interface_close(int xc_handle)
{
    if (libxc_trace)
	fprintf(stderr, "libxc: %s\n", __FUNCTION__);
    return 0;
}

int xc_domain_getinfo(int xc_handle,
                      uint32_t first_domid,
                      unsigned int max_doms,
                      xc_dominfo_t *info)
{
    char filename[256];
    int i, dom, fd, count = 0;

    dom = first_domid;
    for (i = 0; i < max_doms; i++) {

	/* most stupid way to enumerate domains ... */
	for (; dom < DOMID_FIRST_RESERVED; dom++) {
	    snprintf(filename, sizeof(filename), "/var/run/xenner/vmcore.%d", dom);
	    fd = open(filename, O_RDWR);
	    if (-1 == fd)
		continue;
	    break;
	}
	if (-1 == fd)
	    break;
	if (libxc_trace)
	    fprintf(stderr, "libxc: %s: domid %d\n", __FUNCTION__, dom);

	/* good enougth to make the "domain exists?" style users happy */
	info[i].domid = dom;

	close(fd);
	count++;
	dom++;
    }
    return count;
}

/* ------------------------------------------------------------------ */

void *xc_map_foreign_range(int xc_handle, uint32_t dom,
			   int size, int prot,
			   unsigned long mfn)
{
    char filename[256];
    void *ptr;
    int fd;

    snprintf(filename, sizeof(filename), "/var/run/xenner/vmcore.%d", dom);
    fd = open(filename, O_RDWR);
    if (-1 == fd)
	return NULL;
    ptr = mmap(NULL, size, prot, MAP_SHARED, fd,
	       VMCORE_HEADER + mfn * PAGE_SIZE);
    if ((void*)-1 == ptr) {
	fprintf(stderr, "libxc: mmap(%s): %s\n", filename, strerror(errno));
	close(fd);
	return NULL;
    }
    close(fd);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, dom %d, size %d, prot %d, mfn 0x%lx)"
		" -> %p\n",
		__FUNCTION__, xc_handle, dom, size, prot, mfn, ptr);
    return ptr;
}

/* ------------------------------------------------------------------ */

static int fake_xsd_port(void)
{
    char file[32] = "/tmp/xsd_port.XXXXXXXX";
    char buf[16];
    int fd;

    xsd_port = 69;
    if (libxc_trace)
	fprintf(stderr, "libxc: %s, port %d\n", __FUNCTION__, xsd_port);
    snprintf(buf, sizeof(buf), "%d\n", xsd_port);
    fd = mkstemp(file);
    write(fd, buf, strlen(buf));
    lseek(fd, 0, SEEK_SET);
    unlink(file);
    return fd;
}

static int fake_xsd_kva(void)
{
    char file[32] = "/tmp/xsd_kva.XXXXXXXX";

    if (libxc_trace)
	fprintf(stderr, "libxc: %s\n", __FUNCTION__);
    xsd_fd = mkstemp(file);
    lseek(xsd_fd, PAGE_SIZE-1, SEEK_SET);
    write(xsd_fd, '\0', 1);
    lseek(xsd_fd, 0, SEEK_SET);
    xsd_map = mmap(NULL, PAGE_SIZE, PROT_READ | PROT_WRITE,
		   MAP_SHARED, xsd_fd, 0);
    unlink(file);
    return xsd_fd;
}

/* open syscall wrapper, to fake /proc/xen files */
int open(const char *file, int flags, ...)
{
    static int (*real_open)(const char *file, int flags, ...);
    va_list args;
    mode_t mode;

    if (!real_open)
	real_open = dlsym(RTLD_NEXT, "open");

    if (0 == strcmp(file, "/proc/xen/xsd_port"))
	return fake_xsd_port();
    if (0 == strcmp(file, "/proc/xen/xsd_kva"))
	return fake_xsd_kva();

    /* default: forward */
    if (libxc_trace)
	fprintf(stderr, "libxc: %s: %s\n", __FUNCTION__, file);

    va_start(args, flags);
    mode = va_arg(args, mode_t);
    va_end(args);
    return real_open(file, flags, mode);
}
